"""Collection of tools to make new elementary operations
"""

from functools import wraps
from autodiff.structures import Number, Array
import numpy as np


def elementary(deriv_func):
    """Decorator to create an elementary operation

    This takes as an argument a function that calculates the derivative of the function
    the user is calculating. When the decorated function is called, `@elementary` also calls
    `deriv_func` and stores both the value and derivative in a new `Number()` object

    Example:
        >>> import numpy as np
        >>> def sin_deriv(x):
        ...     return {x: np.cos(x.val) * x.deriv[x]}
        >>> @elementary(sin_deriv)
        ... def sin(x):
        ...     return np.sin(x.val)
        >>> a = Number(np.pi / 2)
        >>> sina = sin(a)
        >>> sina.val
        1.0
        >>> sina.deriv[a]
        0.0

    
    Args:
        deriv_func (function): Function specifying the derivative of this function. Must return a dictionary where each key-value pair is the partial derivative of the decorated function
    
    Returns:
        function: Decorated function
    """

    def inner(func):
        @wraps(func)
        def inner_func(*args, **kwargs):
            # Check if args[0] has len. If so, apply the function elementwise and return an array
            # rather than a Number
            try:
                value = func(*args, **kwargs)
                deriv = deriv_func(*args, **kwargs)
                return Number(value, deriv)

            except AttributeError:

                vals = [func(element, *args[1:], **kwargs) for element in args[0]]
                derivs = [deriv_func(element, *args[1:], **kwargs) for element in args[0]]
                numbers = [Number(val, deriv) for val, deriv in zip(vals, derivs)]
                return Array(numbers)


        return inner_func
    return inner

def add_deriv(x,y):
    """Derivative of additions, one of x and y has to be a Number object
    
    Args:
        x: a Number
        y: a Number object or an int/float to be added
    
    Returns:
        The derivative of the sum of x and y
    """
    try:
        d={}
        for key in x._deriv.keys():
            if key in y._deriv.keys():
                d[key] = x._deriv[key] + y._deriv[key]
            else:
                d[key] = x._deriv[key]
        for key in y._deriv.keys():
            if not key in x._deriv.keys():
                d[key] = y._deriv[key]
    except AttributeError:
        d = x._deriv
    return d

@elementary(add_deriv)
def add(x,y):
    """add two numbers together, one of x and y has to be a Number object
    
    Args:
        x: a Number object or an int/float
        y: a Number object or an int/float to be added
    
    Returns:
        value of the sum
    """
    try:
        s = x.val + y.val
    except:
        s = x.val + y
    return s

def subtract_deriv(x,y):
    """Derivative of subtractions, one of x and y has to be a Number object
    
    Args:
        x: a Number
        y: a Number object or an int/float to be subtracted
    
    Returns:
        The derivative of the difference of x and y
    """
    try:
        d={}
        for key in x._deriv.keys():
            if key in y._deriv.keys():
                d[key] = x._deriv[key] - y._deriv[key]
            else:
                d[key] = x._deriv[key]
        for key in y._deriv.keys():
            if not key in x._deriv.keys():
                d[key] = -y._deriv[key]
    except AttributeError:
        d = x._deriv
    return d

@elementary(subtract_deriv)
def subtract(x,y):
    """Subtract one number from another, one of x and y has to be a Number object
    
    Args:
        x: a Number object
        y: a Number object or an int/float to be subtracted
    
    Returns:
        value of the difference
    """
    try:
        s = x.val - y.val
    except:
        s = x.val - y
    return s

def mul_deriv(x,y):
    """Derivative of multiplication, one of x and y has to be a Number object
    
    Args:
        x: a Number
        y: a Number object or an int/float to be multiplied
    
    Returns:
        The derivative of the product of x and y
    """
    if (x==y):
        return pow_deriv(x,2)
    try:
        d={}
        for key in x._deriv.keys():
            if key in y._deriv.keys():
                #product rule
                d[key] = x._deriv[key] * y.val + y._deriv[key] * x.val
            else:
                d[key] = x._deriv[key] * y.val
        for key in y._deriv.keys():
            if not key in x._deriv.keys():
                d[key] = y._deriv[key] * x.val
    except AttributeError:
        d = {}
        for key in x._deriv.keys():
            d[key] = x._deriv[key]*y
    return d

@elementary(mul_deriv)
def mul(x,y):
    """Subtract one number from another, one of x and y has to be a Number object
    
    Args:
        x: a Number object
        y: a Number object or an int/float to be subtracted
    
    Returns:
        value of the difference
    """
    try:
        s = x.val * y.val
    except AttributeError:
        s = x.val * y
    return s

def div_deriv(x,y):
    """Derivative of division, one of x and y has to be a Number object
    
    Args:
        x: a Number
        y: a Number object or an int/float to be divided
    
    Returns:
        The derivative of the quotient of x and y
    """
    try:
        d={}
        for key in x._deriv.keys():
            if key in y._deriv.keys():
                #quotient rule
                d[key] = (-y._deriv[key] * x.val + x._deriv[key] * y.val)/(y.val**2)
            else:
                d[key] = x._deriv[key] / y.val
        for key in y._deriv.keys():
            if not key in x._deriv.keys():
                d[key] = -x.val / (y.val ** 2) * y._deriv[key]
    except AttributeError:
        d = {}
        for key in x._deriv.keys():
            d[key] = x._deriv[key] / y
    return d

@elementary(div_deriv)
def div(x,y):
    """Subtract one number from another, one of x and y has to be a Number object
    
    Args:
        x: a Number object
        y: a Number object or an int/float to be subtracted
    
    Returns:
        value of the difference
    """
    try:
        s = x.val / y.val
    except AttributeError:
        s = x.val / y
    return s


def pow_deriv(x, a):
    """Derivative of power of a Number
    
    Args:
        x: a Number
        a: a Number
    
    Returns:
        The derivative of the power
    """
    d = {}

    # All the derivates w.r.t x
    try:
        for key in x._deriv.keys():
            if key in a._deriv.keys():
                # Using the chain rule for powers
                d[key] = (((a.val * x._deriv[key]) / x.val) + (a._deriv[key] * np.log(x.val))) * (x.val ** a.val)
            else:
                d[key] = a.val * x.val ** (a.val - 1) * x._deriv[key]
    except AttributeError:
        try:
            for key in x._deriv.keys():
                d[key] = a * x.val ** (a - 1) * x._deriv[key]
        except AttributeError:
            # x isn't a Number(). Just go through
            pass

    # All the derivatives w.r.t a
    try:
        for key in a._deriv.keys():
            if key not in x._deriv.keys():
                d[key] = x.val ** a.val * np.log(x.val) * a._deriv[key]
    except AttributeError:
        try:
            for key in a._deriv.keys():
                d[key] = x ** a.val * np.log(x) * a._deriv[key]
        except AttributeError:
            # a isn't a Number(). Just go through
            pass

    return d

@elementary(pow_deriv)
def power(x,y):
    """power of one number by another, one of x and y has to be a Number object
    
    Args:
        x: a Number object or an int/float to be powered
        y: a Number object or an int/float to be the exponential
    
    Returns:
        value of the difference
    """
    # return x.val**y
    try:
        return x.val ** y.val
    except AttributeError:
        try:
            return x ** y.val
        except AttributeError:
            return x.val ** y
            # except AttributeError:
            #     return x ** y

def sin_deriv(x):
    """Derivative of sin(x)
    
    Args:
        x (structures.Number()): Number to take the sin of. Must have a ``deriv`` attribute
    
    Returns:
        dict: dictionary of partial derivatives
    """
    d={}
    for key in x._deriv.keys():
        d[key] = np.cos(x.val) * x._deriv[key]
    d[x] = np.cos(x.val) * x._deriv[x]
    return d

@elementary(sin_deriv)
def sin(x):
    """Take the sin(x)
    
    Args:
        x (Number): Number to take the sin of
    
    Returns:
        float: sin(x.val)
    """
    return np.sin(x.val)

def cos_deriv(x):
    """Derivative of cos(x)
    
    Args:
        x (structures.Number()): Number to take the cos of. Must have a ``deriv`` attribute
    
    Returns:
        dict: dictionary of partial derivatives
    """
    d={}
    for key in x._deriv.keys():
        d[key] = -np.sin(x.val) * x._deriv[key]
    d[x] = -np.sin(x.val) * x._deriv[x]
    return d

@elementary(cos_deriv)
def cos(x):
    """Take the cos(x)
    
    Args:
        x (Number): Number to take the cos of
    
    Returns:
        float: cos(x.val)
    """
    return np.cos(x.val)

def tan_deriv(x):
    """Derivative of tan(x)

    Args: 
        x (structures.Numbers()): Number to take the tan of. Must have a ``deriv`` attribute
    
    Returns:
        dict: dictionary of partial derivatives
    """
    d={}
    for key in x._deriv.keys():
        d[key] = (np.tan(x.val)**2 + 1) * x._deriv[key]
    d[x] = (np.tan(x.val)**2 + 1) * x._deriv[x]
    return d

@elementary(tan_deriv)
def tan(x):
    """Take the tan(x)

    Args:
        x (Number): Number to take the tan of

    Returns:
        float: tan(x.val)
    """
    return np.tan(x.val)

def exp_deriv(x):
    """Derivative of exp(x)
    
    Args:
        x (structures.Number()): Number to take the exp of. Must have a ``deriv`` attribute
    
    Returns:
        dict: dictionary of partial derivatives
    """
    d={}
    for key in x._deriv.keys():
        d[key] = np.exp(x.val) * x._deriv[key]
    d[x] = np.exp(x.val) * x._deriv[x]
    return d

@elementary(exp_deriv)
def exp(x):
    """Take the exp(x)
    
    Args:
        x (Number): Number to take the exp of
    
    Returns:
        float: exp(x.val)
    """
    return np.exp(x.val)

def log_deriv(x, y=np.exp(1)):
    """Derivative of log(x) at base y
    
    Args:
        x (structures.Number()): Number to take the log of. Must have a ``deriv`` attribute
        y (a Number object or an int/float): Base of the logarithm.
    
    Returns:
        dict: dictionary of partial derivatives
    """
    d = {}
    # Use the chain rule to find partials w.r.t everything x depends on
    for key in x._deriv.keys():
        d[key] = 1 / (x.val) * x._deriv[key]

    try:
        d[x] = 1 / (x.val * np.log(y.val))
    except AttributeError:
        d[x] = 1 / (x.val * np.log(y))

    return d

@elementary(log_deriv)
def log(x, y=np.exp(1)):
    """Take the log(x) at base y
    
    Args:
        x (Number): Number to take the log of
        y (a Number object or an int/float): Base of the logarithm.
    
    Returns:
        float: log(x.val)
    """
    try:
        s = np.log(x.val) / np.log(y.val)
    except AttributeError:
        s = np.log(x.val) / np.log(y)
    return s

def negate_deriv(x):
    '''Derivatives of the elementary negation function
    
    Args:
        x (Number): Number to be negated
        
    Returns:
        dictionary: the partial derivatives of the negated Number
    '''
    return {key: -deriv for key, deriv in x._deriv.items()}

@elementary(negate_deriv)
def negate(x):
    """Negate
    
    Args:
        x (Number): Negate a number
    
    Returns:
        Number: -x
    """
    return - x.val

def logistic_deriv(x):
    """Derivative of logistic(x)
    
    Args:
        x (structures.Number()): Number to the natural exponential in the logistic function.
        Must have a ``deriv`` attribute
    
    Returns:
        dict: dictionary of partial derivatives
    """
    d = {}
    for key in x._deriv.keys():
        d[key] = -(1+np.exp(-x.val))**-2*(-np.exp(-x.val)) * x._deriv[key]
    d[x] = -(1+np.exp(-x.val))**-2*(-np.exp(-x.val)) * x._deriv[x]

    return d


@elementary(logistic_deriv)
def logistic(x):
    """Take the logistic(x) 
    
    Args:
        x (Number): Number to the natural exponential in the logistic
    
    Returns:
        float: logistic(x.val)
    """
    return 1.0/(1.0+np.exp(-x.val))


def asin_deriv(x):
    """Arcsin derivative
    
    Args:
        x (Number): Value
    
    Returns:
        dict: Partial derivatives w.r.t. everything x had a partial w.r.t.
    """
    d = {}

    for key in x._deriv.keys():
        d[key] = 1 / np.sqrt(-x.val ** 2 + 1) * x._deriv[key]

    # return 1 / np.sqrt(-x ** 2 + 1)
    return d

@elementary(asin_deriv)
def asin(x):
    """Arcsin
    
    Args:
        x (Number): Value
    
    Returns:
        Number: asin(x)
    """
    return np.arcsin(x.val)

def acos_deriv(x):
    """Derivative of arccos
    
    Args:
        x (Number): Value
    
    Returns:
        dict: Partial derivatives w.r.t. everything x had a partial w.r.t.
    """
    d = {}
    for key in x._deriv.keys():
        d[key] = -1 / np.sqrt(-x.val ** 2 + 1) * x._deriv[key]

    return d

@elementary(acos_deriv)
def acos(x):
    """Arccos
    
    Args:
        x (Number): Value
    
    Returns:
        Number: acos(x)
    """
    return np.arccos(x.val)

def atan_deriv(x):
    """Derivative of atan
    
    Args:
        x (Number): Value
    
    Returns:
        dict: Partial derivatives w.r.t. everything x had a partial w.r.t.
    """
    d = {}
    for key in x._deriv.keys():
        d[key] = 1 / (x.val ** 2 + 1) * x._deriv[key]
    return d

@elementary(atan_deriv)
def atan(x):
    """Arctan
    
    Args:
        x (Number): Value
    
    Returns:
        Number: atan(x)
    """
    return np.arctan(x.val)

def cosh_deriv(x):
    """Hyperbolic cosine
    
    Args:
        x (Number): Value
    
    Returns:
        dict: Partial derivatives w.r.t. everything x had a partial w.r.t.
    """
    d = {}
    for key in x._deriv.keys():
        d[key] = np.sinh(x.val) * x._deriv[key]
    return d

@elementary(cosh_deriv)
def cosh(x):
    """Hyperbolic cosine
    
    Args:
        x (Number): Value
    
    Returns:
        Number: cosh(x)
    """
    return np.cosh(x.val)

def sinh_deriv(x):
    """Hyperbolic sin derivative
    
    Args:
        x (Number): Value
    
    Returns:
        dict: Partial derivatives w.r.t. everything x had a partial w.r.t.
    """
    d = {}
    for key in x._deriv.keys():
        d[key] = np.cosh(x.val) * x._deriv[key]
    return d

@elementary(sinh_deriv)
def sinh(x):
    """Hyperbolic sine
    
    Args:
        x (Number): Value
    
    Returns:
        Number: sinh(x)
    """
    return np.sinh(x.val)

def tanh_deriv(x):
    """Hyperbolic tan derivative
    
    Args:
        x (Number): Value
    
    Returns:
        dict: Partial derivatives w.r.t. everything x had a partial w.r.t.
    """
    d = {}
    for key in x._deriv.keys():
        d[key] = -np.tanh(x.val) ** 2 + 1
    return d

@elementary(tanh_deriv)
def tanh(x):
    """Hyperbolic tan
    
    Args:
        x (Number): Value
    
    Returns:
        Number: tanh(x)
    """
    return np.tanh(x.val)

def sqrt_deriv(x):
    '''Derivative of the square root function
    
    Args:
        x(Number): Number to take the square root of
    
    Returns:
        dictionary: the partial derivatives of the square root of number
    '''
    d = {}
    for key in x._deriv.keys():
        # d[key] = (1 / 2) * 1 / np.sqrt(x.deriv[key])
        d[key] = 1 / (2 * np.sqrt(x.val)) * x._deriv[key]

    return d

@elementary(sqrt_deriv)
def sqrt(x):
    '''Square root of a number
    
    Args:
        x(Number): Number to take the square root of
        
    Returns:
        float: the square root of Number x's value
    '''
    return np.sqrt(x.val)




"""Utilities for writing code
"""

import sympy as sp
x = sp.symbols('x')
def f_and_deriv(f, wrt=x):
    sp.printing.pprint('Function:')
    sp.printing.pprint(f)
    print('\n', sp.printing.pycode(f))
    sp.printing.pprint('\nDeriv:')
    sp.printing.pprint(f.diff(wrt))
    print('\n', sp.printing.pycode(f.diff(wrt)))

if __name__ == '__main__':
    f_and_deriv(sp.asin(x))


"""Data structures for autodiff
"""
from autodiff import operations
import numpy as np

class Number():
    '''
    Number class is the core data structure for 'autodiff'. It instantiates a Number 
    object by specifying a value and a derivative. Derivative with respect to itself 
    is automatically instantiated to 1.
    
    Args:
        val: value of the Number
        deriv: a dictionary of partial derivatives. It is automatically instantiated to
            {self: 1} unless otherwise specified.
    
    Returns:
        Number, an object to perform automatic differentiation on.
        
    Example:
        >>> import autodiff
        >>> x = autodiff.structures.Number(3)
        >>> x.value
        3
        >>> x.jacobian(x)
        1
        >>> a = autodiff.structures.Number(3,2)
        >>> a.value
        3
        >>> a.jacobian(x)
        2
    '''
    
    def __init__(self, val, deriv=None):

        self.val = val
        if deriv is None:
            self._deriv = {
                self: 1
            }
        elif isinstance(deriv, dict):
            self._deriv = deriv
            #keep also a copy of the derivative w.r.t. itself
            self._deriv[self] = 1
        else:
            self._deriv = {
                    self: deriv
                    }

    def __repr__(self):
        '''
        Overloads the print method to give a string representation of Number object.
        
        Returns:
            a string specifiying the value of the Number object.
        '''
        return f'Number(val={self.val})'
    
    def __add__(self, other):
        '''
        Overloads the add method to add a number object to another Number object or an integer/float.
        
        Args:
            other: a Number object or an integer or float to be added.
        
        Returns:
            another Number object, which is the sum.
        '''
        return operations.add(self, other)
    
    def __radd__(self, other):
        '''
        Overloads the right add method to add a number object to another Number object or an integer/float.
        
        Args:
            other: a Number object or an integer or float to be added.
        
        Returns:
            another Number object, which is the sum.
        '''
        return operations.add(self, other)
    
    def __sub__(self, other):
        '''
        Overloads the subtract method to subtract from a number object by another Number object or an integer/float.
        
        Args:
            other: a Number object or an integer or float to be subtracted
        
        Returns:
            another Number object, which is the difference
        '''
        return operations.subtract(self, other)
    
    def __rsub__(self, other):
        '''
        Overloads the right subtract method to subtract from a number object by another Number object or an integer/float.
        
        Args:
            other: a Number object or an integer or float to be subtracted
        
        Returns:
            another Number object, which is the difference
        '''
        return -operations.subtract(self, other)

    def __mul__(self, other):
        '''
        Overloads the multiply method to multiply a number object by another Number object or an integer/float.
        
        Args:
            other: a Number object or an integer or float to be multiplied
        
        Returns:
            another Number object, which is the product
        '''
        return operations.mul(self, other)
    
    def __rmul__(self, other):
        '''
        Overloads the right multiply method to multiply a number object by another Number object or an integer/float.
        
        Args:
            other: a Number object or an integer or float to be multiplied
        
        Returns:
            another Number object, which is the product
        '''
        return operations.mul(self, other)
    
    def __truediv__(self, other):
        '''
        Overloads the division method to divide a number object by another Number object or an integer/float.
        
        Args:
            other: a Number object or an integer or float to be divided
        
        Returns:
            another Number object, which is the quotient
        '''
        return operations.div(self, other)

    def __rtruediv__(self, other):
        '''
        Overloads the right division method to divide a number object by another Number object or an integer/float.
        
        Args:
            other: a Number object or an integer or float to be divided
        
        Returns:
            another Number object, which is the quotient
        '''
        return operations.div(self, other) ** -1
    
    def __pow__(self, other):
        '''
        Overloads the power method to power a number object by another Number object or an integer/float.
        
        Args:
            other: a Number object or an integer or float to be the power
        
        Returns:
            another Number object, which is the original number ** other
        '''
        return operations.power(self, other)

    def __rpow__(self, other):
        '''
        Overloads the right power method to power a number object by another Number object or an integer/float.
        
        Args:
            other: a Number object or an integer or float to be the power
        
        Returns:
            another Number object, which is the original number ** other
        '''
        return operations.power(other, self)

    def __neg__(self):
        '''
        Overloads the negation method to negate a number object
        
        Returns:
            another Number object, which is the negation of the original number
        '''
        return operations.negate(self)
    
    # def sin(self):
    #     '''
    #     Calculates the sin of the Number object.
        
    #     Returns:
    #         another Number object, which is sin of the original one.
    #     '''
    #     return operations.sin(self)
    
    # def asin(self):
    #     '''
    #     Calculates the arcsin of the Number object.
        
    #     Returns:
    #         another Number object, which is arcsin of the original one.
    #     '''
    #     return operations.asin(self)
    
    # def sinh(self):
    #     '''
    #     Calculates the sinh of the Number object.
        
    #     Returns:
    #         another Number object, which is sinh of the original one.
    #     '''
    #     return operations.sinh(self)
    
    # def cos(self):
    #     '''
    #     Calculates the cosine of the Number object.
        
    #     Returns:
    #         another Number object, which is cosine of the original one.
    #     '''
    #     return operations.cos(self)
    
    # def acos(self):
    #     '''
    #     Calculates the arccosine of the Number object.
        
    #     Returns:
    #         another Number object, which is arccosine of the original one.
    #     '''
    #     return operations.acos(self)
    
    # def cosh(self):
    #     '''
    #     Calculates the cosine-h of the Number object.
        
    #     Returns:
    #         another Number object, which is cosine-h of the original one.
    #     '''
    #     return operations.cosh(self)
    
    # def tan(self):
    #     '''
    #     Calculates the tangent of the Number object.
        
    #     Returns:
    #         another Number object, which is tangent of the original one.
    #     '''
    #     return operations.tan(self)
    
    # def atan(self):
    #     '''
    #     Calculates the arc-tangent of the Number object.
        
    #     Returns:
    #         another Number object, which is arc-tangent of the original one.
    #     '''
    #     return operations.atan(self)
    
    # def tanh(self):
    #     '''
    #     Calculates the tangent-h of the Number object.
        
    #     Returns:
    #         another Number object, which is tangent-h of the original one.
    #     '''
    #     return operations.tanh(self)
    
    # def sqrt(self):
    #     '''
    #     calculates the square root of the Number object.
        
    #     Returns:
    #         another number object, which is the square root of the original one.
    #     '''
    #     return operations.sqrt(self)
    

    # def exp(self):
    #     '''
    #     Calculates the exponential of Number object.
        
    #     Returns:
    #         another Number object, which is the exponential of the original one.
    #     '''
    #     return operations.exp(self)

    # def logistic(self):
        
    #     '''
    #     Calculates the logistic of Number object.
        
    #     Returns:
    #         another Number object, which is the logistic of the original one.
    #     '''

    #     return operations.logistic(self)
    
    # def log(self, base = np.exp(1)):
    #     '''
    #     Calculates the log of Number object.
        
    #     Args:
    #         base: the base to take log with
        
    #     Returns:
    #         another Number object, which is the log of the original one.
    #     '''

    #     return operations.log(self, base)

    def jacobian(self, order):
        '''
        Returns the jacobian matrix by the order specified.
        
        Args:
            order: the order to return the jacobian matrix in. Has to be not null
        
        Returns:
            a np.ndarray of partial derivatives specified by the order.
            When order is a single element, it returns a scaler
        '''
        
        def _partial(deriv, key):
            try:
                return deriv[key]
            except KeyError:
                # If there's no partial, it's zero
                return 0
        
        #if order is a single Number object
        if isinstance(order, Number):
            return _partial(self._deriv, order)
        
        jacobian = []
        for key in order:
            jacobian.append(_partial(self._deriv, key))
        return np.array(jacobian)

    def __hash__(self):
        return id(self)
  
    def __eq__(self, other):
        '''
        Overloads the Comparison Operator to check whether two Number 
        objects are equal to each other
   
        Args:
            other: the other Number object to be compared with
   
        Returns:
            True if two Number objects are equal, False otherwise.
        '''
        try:
            if self.val == other.val:
                deriv_self = self._deriv.copy()
                deriv_other = other._deriv.copy()
                deriv_self.pop(self)
                deriv_other.pop(other)
                if deriv_self==deriv_other:
                    return True
            return False
        except AttributeError:
            #if other is not even a autodiff.Number
            return False
   
    def __ne__(self, other):
        '''
        Overloads the Comparison Operator to check whether two Number 
        objects are not equal to each other
   
        Args:
            other: the other Number object to be compared with
   
        Returns:
            True if two Number objects are not equal, False otherwise.
        '''
        return not self.__eq__(other)

class Array():
    '''
    Array class is another core data structure for 'autodiff'. It instantiates an Array 
    object as a list of Number objects. It holds the list internally as an np.ndarray
    
    Args:
        iterable: a iterable objects of Numbers to be held in Array. If lists of 
        ints/floats are passed, it will convert them to Number objects first 
    
    Returns:
        Array, an object to perform automatic differentiation on.
        
    Example:
        >>> import autodiff
        >>> x = autodiff.structures.Number(3)
        >>> a = autodiff.structures.Number(3,2)
        >>> arr = autodiff.structures.Array([x, a])
        Array([Number(val=3) Number(val=3)])
        >>> arr.jacobian(a)
        array([0, 2])
    '''

    def __init__(self, iterable):
        self._data = np.array(iterable, dtype=np.object)
        for i, d in enumerate(self._data):
            if not isinstance(d, Number):
                self._data[i] = Number(d)

    def __str__(self):
        '''
        Overloads the string method to give a string representation of Array object.
        
        Returns:
            a string specifiying the value of each Number object in Array.
        '''
        return str(self._data)

    def __repr__(self):
        '''
        Overloads the print method to give a string representation of Array object.
        
        Returns:
            a string specifiying the value of each Number object in Array.
        '''
        return f'Array({self._data})'

    def __len__(self):
        '''
        Overloads the len() method to give the length of Array.
        
        Returns:
            an integer representing number of elements in Array.
        '''
        return len(self._data)

    def __getitem__(self, idx):
        '''
        Overloads [idx] to get an item of the Array at a specific postion.
        
        Args:
            idx, the index to take element at.
        
        Returns:
            a Number object, which is at position 'idx' of the Array.
        '''
        return self._data[idx]

    def __setitem__(self, idx, val):
        '''
        Sets an item of the Array at a specific postion with the value specified.
        
        Args:
            idx, the index to set element at
            val, the Number object to put in Array
        '''
        if isinstance(val, Number):
            self._data[idx] = val
        else:
            raise ValueError('invalid literal for Number(): {}'.format(val))

    def __add__(self, other):
        '''
        Overloads addition to add two arrays, or a Number/int/float to an array.
        
        Args:
            other, another Array of same length to perform element-wise addition on,
            or a Number object to be added, or an integer/float to be added.
        
        Returns:
            an Array object, which is the sum.
        '''
        try:
            # If they're both Array2 objects
            return Array(self._data.__add__(other._data))
        except AttributeError:
            return Array(self._data.__add__(other))

    def __radd__(self, other):
        '''
        Overloads right addition to add two arrays, or a Number/int/float to an array.
        
        Args:
            other, another Array of same length to perform element-wise addition on,
            or a Number object to be added, or an integer/float to be added.
        
        Returns:
            an Array object, which is the sum.
        '''
        return self._data.__radd__(other)

    def __sub__(self, other):
        '''
        Overloads subtraction to subtract two arrays, or a Number/int/float from an array.
        
        Args:
            other, another Array of same length to perform element-wise subtraction on,
            or a Number object to be subtracted, or an integer/float to be subtracted.
        
        Returns:
            an Array object, which is the difference.
        '''
        try:
            # If they're both Array2 objects
            return Array(self._data.__sub__(other._data))
        except AttributeError:
            return Array(self._data.__sub__(other))

    def __rsub__(self, other):
        '''
        Overloads right subtraction to subtract two arrays, or a Number/int/float 
        from an array.
        
        Args:
            other, another Array of same length to perform element-wise subtraction on,
            or a Number object to be subtracted, or an integer/float to be subtracted.
        
        Returns:
            an Array object, which is the difference.
        '''
        return self._data.__rsub__(other)

    def __mul__(self, other):
        '''
        Overloads multiplication to multiply two arrays, or a Number/int/float to an array.
        
        Args:
            other, another Array of same length to perform element-wise multiplication on,
            or a Number object to be multiplied, or an integer/float to be multiplied.
        
        Returns:
            an Array object, which is the product.
        '''
        try:
            # If they're both Array2 objects
            return Array(self._data.__mul__(other._data))
        except AttributeError:
            return Array(self._data.__mul__(other))

    def __rmul__(self, other):
        '''
        Overloads right multiplication to multiply two arrays, or a Number/int/float 
        to an array.
        
        Args:
            other, another Array of same length to perform element-wise multiplication on,
            or a Number object to be multiplied, or an integer/float to be multiplied.
        
        Returns:
            an Array object, which is the product.
        '''
        return self._data.__rmul__(other)

    def __truediv__(self, other):
        '''
        Overloads division to divide two arrays, or a Number/int/float from an array.
        
        Args:
            other, another Array of same length to perform element-wise division on,
            or a Number object to divide by, or an integer/float to divide by.
        
        Returns:
            an Array object, which is the quotient.
        '''
        try:
            # If they're both Array2 objects
            return Array(self._data.__truediv__(other._data))
        except AttributeError:
            return Array(self._data.__truediv__(other))

    def __rtruediv__(self, other):
        '''
        Overloads right division to divide two arrays, or a Number/int/float from an array.
        
        Args:
            other, another Array of same length to perform element-wise division on,
            or a Number object to divide by, or an integer/float to divide by.
        
        Returns:
            an Array object, which is the quotient.
        '''
        return self._data.__rtruediv__(other)

    def __matmul__(self, other):
        '''
        Overloads matrix multiplication to multiply two matrices, 
        or a Number/int/float to a matrix.
        
        Args:
            other, another matrix to be multiplied,
            or a Number object to be multiplied, or an integer/float to be multiplied.
        
        Returns:
            an Array object, which is the product.
        '''
        try:
            # If they're both Array objects
            out = self._data.__matmul__(other._data)
        except AttributeError:
            out = self._data.__matmul__(other)
        return(out)

    def __rmatmul__(self, other):
        '''
        Overloads right matrix multiplication to multiply two matrices, 
        or a Number/int/float to a matrix.
        
        Args:
            other, another matrix to be multiplied,
            or a Number object to be multiplied, or an integer/float to be multiplied.
        
        Returns:
            an Array object, which is the product.
        '''
        return self._data.__rmatmul__(other)

    def __pow__(self, other):
        '''
        Overloads power to take the power of an Array object.
        
        Args:
            other, another Array of same dimension to perform element-wise power by,
            or a Number object to be take the power of, or an integer/float to take
            the power of.
        
        Returns:
            an Array object, which is the power.
        '''
        try:
            # If they're both Array2 objects
            return Array(self._data.__pow__(other._data))
        except AttributeError:
            return Array(self._data.__pow__(other))

    def __rpow__(self, other):
        '''
        Overloads right power to take the power of an Array object.
        
        Args:
            other, another Array of same dimension to perform element-wise power by,
            or a Number object to be take the power of, or an integer/float to take
            the power of.
        
        Returns:
            an Array object, which is the power.
        '''
        return self._data.__rpow__(other)

    def __neg__(self):
        '''
        Overloads negation on Array objects.
        
        Returns:
            another Array object, which is the negated original Array.
        '''
        return Array(self._data.__neg__())

    def jacobian(self, order):
        '''
        Returns the jacobian matrix by the order specified.
        
        Args:
            order: the order to return the jacobian matrix in. Has to be not null
        
        Returns:
            a np.ndarray of partial derivatives specified by the order. Each row is
            an element in the original array, each column is the order specified.
            When order is a single element, it returns a flat array.
        '''
        def _partial(deriv, key):
            try:
                return deriv[key]
            except KeyError:
                # If there's no partial, it's zero
                return 0

        jacobian = []
        for element in self._data:
            # Check if order is iterable
            try:
                jacobian_row = []
                for key in order:
                    jacobian_row.append(_partial(element._deriv, key))
                jacobian.append(jacobian_row)
            except TypeError:
                jacobian.append(_partial(element._deriv, order))
        
        return np.array(jacobian)
    
    def __eq__(self, other):
        '''
        Overloads the Comparison Operator to check whether two Array 
        objects are equal to each other
   
        Args:
            other: the other Array object to be compared with
   
        Returns:
            True if two Array objects are equal, False otherwise.
        '''
        try:
            initial = True
            if len(self)==len(other):
                for i in range(len(self)):
                    if not (self[i]==other[i]):
                        initial = False
                return initial
            else:
                return False

        except TypeError:
            #if other is not even a autodiff.Number
            return False
   
    def __ne__(self, other):
        '''
        Overloads the Comparison Operator to check whether two Array 
        objects are not equal to each other
   
        Args:
            other: the other Array object to be compared with
   
        Returns:
            True if two Array objects are not equal, False otherwise.
        '''
        return not self.__eq__(other)




from autodiff import operations
from autodiff.structures import Number
from autodiff.structures import Array
import numpy as np
from copy import deepcopy

def newtons_method(func, initial_guess, iterations=100,tolerance = 10**-7,verbose = False, show_fxn=False):
    
    """Use Newton's method to find the root of the function
    Args:
        func: the function that the user wants to find root for
        initial_guess: a number object for the initial guess
        iterations: number of maximum iterations
        show_fxn: if true, return function value at the final xstar

    Returns:
        xn: the x value of the root
        jacobians: the jacobian at each step of the root_finding
        fxn: func(xn). If root_finding is successful, this value should be 0
    """    
    if isinstance(initial_guess,Number):
        #scalar case
        jacobians = []

        x0 = initial_guess

        fxn = func(initial_guess)

        fpxn = fxn.jacobian(initial_guess)

        x1 = x0 - fxn/fpxn

        jacobians.append(fpxn)

        for i in range(iterations):
            if abs(fxn.val) > 1e-7:
                x0 = x1

                fxn = func(x0)

                fpxn = fxn.jacobian(x0)

                jacobians.append(fpxn)

                x1 = x0- fxn / fpxn

        if show_fxn:
            return x1, jacobians,fxn
        else:
            return x1, jacobians
    elif isinstance(initial_guess,Array):
        jacobians = []
        if isinstance(func(initial_guess),tuple):
            for i in range(iterations):

                if i == 0:
                    x0= initial_guess
                else:
                    x0 = x1

                fxn = func(x0)
                fpxn = []
                if verbose:
                    print(i,x0,fxn)

                vector = []
                for n in range(len(fxn)):
                    vector.append(fxn[n].val)


                if np.linalg.norm(vector)< tolerance:
                    break

                for k in range(len(fxn)):
                    fpxn_row = []
                    for j in range(len(x0)):
                        fpxn_row.append(fxn[k].jacobian(x0[j]))
                    fpxn.append(fpxn_row)
                jacobians.append(fpxn)
                x1 = x0 - np.dot(np.linalg.inv(fpxn),fxn)

            if show_fxn:
                return x1, jacobians,fxn
            else:
                return x1, jacobians
            
        else:
            for i in range(iterations):

                if i == 0:
                    x0= initial_guess
                else:
                    x0 = x1

                fxn = func(x0)
                fpxn = []
                if verbose:
                    print(i,x0,fxn)

                if abs(fxn.val)< tolerance:
                    break

                for j in range(len(x0)):
                    fpxn.append(fxn.jacobian(x0[j]))
                jacobians.append(fpxn)
                print(fpxn)
                x1 = x0 - np.dot(np.reciprocal(fpxn),fxn)

            if show_fxn:
                return x1, jacobians,fxn
            else:
                return x1, jacobians



import sys
sys.path.append('..')

import autodiff.operations as operations
from autodiff.structures import Number
from autodiff.structures import Array
import numpy as np

def bfgs_symbolic(func,gradient, initial_guess,iterations =100,tolerance=10**-8,verbose=False):
    """Use symbolic BFGS method to find the local minimum/maxinum of the function
    Args:
        func: the function that the user wants to optimize
        gradient: the expression of the gradient of the function
        initial_guess: an array or single value for the initial guess
        iterations: number of maximum iterations
        tolerance: tolerance
        verbose: if True, print the guess at every step

    Returns:
        x0: the x value of the local extremum
        func(x0): the value of the local extremum
        jacobians: the jacobians of each optimization step
        """   
    # if len(initial_guess)==None:
    # if isinstance(initial_guess, Sized):
    try:
        jacobians = []

        for i in range(iterations):

            if i == 0:
                
                x0 = initial_guess
                #initial guess of hessian
                # This will give an error if initial_guess is a scalar
                H = np.identity(len(x0))
            else:
                x0 = x1
                H = deltaH
            if verbose:
                print(i,x0,func(x0))
            fxn0 = np.array(func(x0))
            fpxn0 = np.array(gradient(x0))    
            jacobians.append(fpxn0)
            if np.linalg.norm(fpxn0)<tolerance:
                #optimization condition is met
                break
                
            s = -np.dot(H,fpxn0) #np.array multiply with scalar would be fine
            x1 = x0 + s
            fxn1 = func(x1)
            fpxn1 = gradient(x1)
            y = np.array(fpxn1 - fpxn0)
            rho0 = 1/(np.dot(y.T,s))
            rhokykT = rho0*y.T
            skrhokykT = np.dot(s.reshape([len(x0),1]),rhokykT.reshape([1,len(x0)]))

            ykskT = np.dot(np.reshape(y,[len(x0),1]),s.reshape([len(x0),1]).T)

            rhokykskT = rho0*ykskT

            skskT = np.dot(s.reshape([len(x0),1]),s.reshape([len(x0),1]).T)

            rhokskskT = rho0*skskT

            #define delta H
            deltaH = np.dot((np.identity(2)-skrhokykT),np.dot(H,(np.identity(2)-rhokykskT)))+rhokskskT


        return x0, func(x0), jacobians
    except TypeError:
        # We have a scalar function
    
        x0 = initial_guess
        
        #initial guess of hessian
        b0 = 1
        
        fxn0 = func(x0)

        fpxn0 = gradient(x0)
        
        jacobians = []
        
        jacobians.append(fpxn0)
        

        for i in range(iterations):
            
            if np.abs(fpxn0)<tolerance:
                break
                
            fxn0 = func(x0)

            fpxn0 = gradient(x0)

            s0 = -fpxn0/b0

            x1=x0+s0 
                
            fxn1 = func(x1)
            fpxn1 = gradient(x1)
                
                
            y0 = fpxn1-fpxn0
                
            if y0 == 0:
                break
                    
            delta_b = y0/s0-b0
            b1 = b0 + delta_b
                
            x0 = x1
                
            b0 = b1
                
            jacobians.append(fpxn1)

        return x0,func(x0),jacobians

def bfgs(func, initial_guess,iterations =100,tolerance = 10**-8,verbose = False):
    """Use AD BFGS method to find the local minimum/maxinum of the function
    Args:
        func: the function that the user wants to optimize
        initial_guess: an array or single value for the initial guess
        iterations: number of maximum iterations
        tolerance: tolerance
        verbose: if True, print the guess at every step

    Returns:
        x0: the x value of the local extremum
        func(x0): the value of the local extremum
        jacobians: the jacobians of each optimization step
        """   
    if isinstance(initial_guess,Number): 
    #bfgs for scalar functions
    
        x0 = initial_guess
        
        #initial guess of hessian
        b0 = 1
        
        fxn0 = func(x0)

        fpxn0 = fxn0.jacobian(x0)
        
        jacobians = []
        
        jacobians.append(fpxn0)
        

        for i in range(iterations):
            #stopping criterion
            if np.abs(fpxn0)<tolerance:
                break
                
            fxn0 = func(x0)

            fpxn0 = fxn0.jacobian(x0)

            s0 = -fpxn0/b0

            x1=x0+s0 
                
            fxn1 = func(x1)
            fpxn1 = fxn1.jacobian(x1)
                
                
            y0 = fpxn1-fpxn0
                
            if y0 == 0:
                break
                    
            delta_b = y0/s0-b0
            b1 = b0 + delta_b
                
            x0 = x1
                
            b0 = b1
                
            jacobians.append(fpxn1)

        return x0,func(x0),jacobians

    if isinstance(initial_guess,Array):
        
        jacobians = []

        for i in range(iterations):

            if i == 0:
                
                x0 = initial_guess
                #initial guess of hessian
                H = np.identity(len(x0))
            else:
                x0 = x1
                H = deltaH
            if verbose:
                print(i,x0,func(x0))
                
            fxn0 = func(x0)
            fpxn0 = np.array(fxn0.jacobian(x0))   
            jacobians.append(fpxn0)
            if np.linalg.norm(fpxn0)<tolerance:
                #optimization condition is met
                break
                
            s = -np.dot(H,fpxn0) #np.array multiply with scalar would be fine
            x1 = x0 + s
            fxn1 = func(x1)
            fpxn1 = fxn1.jacobian(x1)
            y = np.array(fpxn1 - fpxn0)
            rho0 = 1/(np.dot(y.T,s))
            rhokykT = rho0*y.T
            skrhokykT = np.dot(s.reshape([len(x0),1]),rhokykT.reshape([1,len(x0)]))

            ykskT = np.dot(np.reshape(y,[len(x0),1]),s.reshape([len(x0),1]).T)

            rhokykskT = rho0*ykskT

            skskT = np.dot(s.reshape([len(x0),1]),s.reshape([len(x0),1]).T)

            rhokskskT = rho0*skskT

            #define delta H
            deltaH = np.dot((np.identity(2)-skrhokykT),np.dot(H,(np.identity(2)-rhokykskT)))+rhokskskT


        return x0,func(x0),jacobians

  

def steepest_descent(func,initial_guess,iterations = 100,step_size=0.01,tolerance = 10**-8,verbose=False):

    """
        Use steepest_descent method to find the local minimum/maxinum of the function
    Args:
        func: the function that the user wants to optimize 
        initial_guess: A number object for the initial guess
        iterations: number of maximum iterations
        step_size: the size of each step

    Returns:
        x0: the x value of the local extremum
        func(x0): the value of the local extremum
        jacobians: the jacobians of each optimization step
    """    
    #gradient descent for scalar functions
    if isinstance(initial_guess,Number):
        x0=initial_guess
        jacobians = []
        s = -func(initial_guess).jacobian(x0)
        jacobians.append(s)
        for i in range(iterations):
            if np.abs(s)>1*10**-7:
                x0 = x0 + step_size*s
                s = -func(x0).jacobian(x0)
                jacobians.append(s)
        
        
        return x0,func(x0),jacobians

    elif isinstance(initial_guess,Array):

        # e.g. R2 --> R1
        jacobians = []
        
        s = np.zeros(len(initial_guess))

        for i in range(iterations):
            if i == 0:
                x0 = initial_guess
            else:
                alpha = step_size
                #perform line search using Armijo condition
                while func(x0+alpha*s).val>func(x0).val+alpha*0.0001*np.dot(np.transpose(-1*s),s):
                    alpha = alpha/2
                
                x0 = x0+alpha*s
            
            if verbose:
                print(i,x0,func(x0))
            for j in range(len(s)):
                s[j] = func(x0).jacobian(x0[j])*-1
            
            if np.abs(s).all()<10**-7:
                break
            jacobians.append(s)
        return i,x0,func(x0),jacobians
